from pygls.workspace import Document

TEST_XSD = """<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
  <xs:element name="testElement" type="CustomComplexType">
    <xs:annotation>
      <xs:documentation xml:lang="en">
        <![CDATA[Documentation ``example``.]]>
      </xs:documentation>
      <xs:documentation xml:lang="es">
        <![CDATA[``Ejemplo`` de documentación.]]>
      </xs:documentation>
    </xs:annotation>
  </xs:element>
  <xs:complexType name="CustomComplexType">
    <xs:sequence>
      <xs:element name="firstElement" type="CustomIntSimpleType" minOccurs="0" maxOccurs="unbounded">
        <xs:annotation>
          <xs:documentation xml:lang="en">
            <![CDATA[
              This is a multiline
              documentation example.
            ]]>
          </xs:documentation>
        </xs:annotation>
      </xs:element>
      <xs:element name="secondElement" type="TestComplexContent"/>
      <xs:element name="thirdElement">
        <xs:complexType>
          <xs:sequence>
            <xs:element name="childElement" type="TestSimpleContent" minOccurs="1" maxOccurs="1"/>
          </xs:sequence>
          <xs:attribute name="testDate" type="xs:date"/>
        </xs:complexType>
      </xs:element>
      <xs:element name="element_with_group" type="ComplexWithGroup"/>
    </xs:sequence>
    <xs:attribute name="id" type="xs:integer" use="required"/>
    <xs:attribute name="value" type="ValueSimpleType"/>
    <xs:attributeGroup ref="TestAttrGroup"/>
  </xs:complexType>
  <xs:complexType name="ComplexWithGroup">
    <xs:sequence>
      <xs:group ref="TestGroup" minOccurs="0" maxOccurs="1"/>
    </xs:sequence>
    <xs:attribute name="group_id" type="xs:integer"/>
  </xs:complexType>
  <xs:complexType name="TestSimpleContent">
    <xs:simpleContent>
      <xs:extension base="xs:string">
        <xs:attribute name="simple" type="ValueSimpleType" use="required"/>
      </xs:extension>
    </xs:simpleContent>
  </xs:complexType>
  <xs:complexType name="TestComplexContent">
    <xs:complexContent>
      <xs:extension base="ComplexWithGroup">
        <xs:sequence>
          <xs:element name="content" type="xs:string"/>
        </xs:sequence>
        <xs:attribute name="complex" type="xs:string"/>
      </xs:extension>
    </xs:complexContent>
  </xs:complexType>
  <xs:simpleType name="CustomIntSimpleType">
    <xs:restriction base="xs:integer">
      <xs:minInclusive value="1"/>
      <xs:maxInclusive value="30"/>
    </xs:restriction>
  </xs:simpleType>
  <xs:simpleType name="ValueSimpleType">
    <xs:restriction base="xs:string">
        <xs:enumeration value="v1"/>
        <xs:enumeration value="v2"/>
        <xs:enumeration value="v3"/>
    </xs:restriction>
  </xs:simpleType>
  <xs:attributeGroup name="TestAttrGroup">
    <xs:attribute name="gattr1" type="xs:string"/>
    <xs:attribute name="gattr2" type="xs:string"/>
  </xs:attributeGroup>
  <xs:group name="TestGroup">
    <xs:sequence>
      <xs:element name="group_elem1" type="xs:string"/>
      <xs:element name="group_elem2" type="xs:string"/>
    </xs:sequence>
  </xs:group>
</xs:schema>
"""
# Resulting XML tree for quick visual reference
# Please update it if the XSD above is modified:
#   print(tree.render())
"""
[testElement] id value gattr1 gattr2
├── [firstElement]
├── [secondElement] group_id complex
│   ├── [group_elem1]
│   ├── [group_elem2]
│   └── [content]
├── [thirdElement] testDate
│   └── [childElement] simple
└── [element_with_group] group_id
    ├── [group_elem1]
    └── [group_elem2]
"""

RECURSIVE_XSD = """<xs:schema xmlns:xs="http://www.w3.org/2001/XMLSchema">
  <xs:element name="node" type="nodeType"/>
  <xs:complexType name="nodeType">
    <xs:sequence>
      <xs:element name="node" type="nodeType"/>
    </xs:sequence>
  </xs:complexType>
</xs:schema>
"""

TEST_TOOL_01 = """<tool id="test" name="Test Tool 01" version="0.1.0">
    <command detect_errors="exit_code"><![CDATA[
        TODO: Fill in command template.
    ]]></command>
    <inputs>
    </inputs>
    <outputs>
    </outputs>
    <help><![CDATA[
        TODO: Fill in help.
    ]]></help>
</tool>
"""
TEST_TOOL_01_DOCUMENT = Document("file://test01.xml", TEST_TOOL_01)

TEST_TOOL_WITH_MACRO_01 = """
<tool id="test_with_macro" name="Test with macro 01" version="@WRAPPER_VERSION@">
    <macros>
        <import>macros.xml</import>
    </macros>
    <expand macro="inputs" />
</tool>"""
TEST_TOOL_WITH_MACRO_01_DOCUMENT = Document("file://test_with_macro_01.xml", TEST_TOOL_WITH_MACRO_01)

TEST_MACRO_01 = """
<macros>
    <token name="@WRAPPER_VERSION@">0.1.0</token>
    <macro name="inputs">
        <inputs/>
    </macro>
</macros>
"""
TEST_MACRO_01_DOCUMENT = Document("file://macros.xml", TEST_MACRO_01)


TEST_INVALID_TOOL_01_DOCUMENT = Document("file://test_invalid_01.xml", "<tool></tool>")

TEST_SYNTAX_ERROR_TOOL_01_DOCUMENT = Document("file://test_syntax_error_01.xml", "tool")

TEST_SYNTAX_ERROR_MACRO_01_DOCUMENT = Document("file://test_syntax_error_macro_01.xml", "<macros>unclosed")



"""Utilities to validate Galaxy xml tool wrappers.
"""

from typing import List, Optional
from lxml import etree
from galaxy.util import xml_macros

from pygls.workspace import Document
from pygls.types import (
    Diagnostic,
    Position,
    Range,
)

MACROS_TAG = "<macros>"


class GalaxyToolValidationService:
    """Service providing diagnostics for errors in the XML validation."""

    def __init__(self, server_name: str, xsd_schema: etree.XMLSchema):
        """Initializes the validator"""
        self.server_name = server_name
        self.xsd_schema = xsd_schema

    def validate_document(self, document: Document) -> List[Diagnostic]:
        """Validates the XML document and returns a list of diagnotics
        if there are any problems.

        Args:
            document (Document): The XML document. Can be a tool wrapper or macro
            definition file.

        Returns:
            List[Diagnostic]: The list of issues found in the document.
        """
        try:
            if self._is_macro_definition_file(document):
                return self._check_syntax(document)

            xml_tree = etree.fromstring(document.source)
            return self._validate_tree(xml_tree)
        except ExpandMacrosFoundException as e:
            result = self._validate_tree_with_macros(document, e.xml_tree)
            return result
        except etree.XMLSyntaxError as e:
            return self._build_diagnostics_from_syntax_error(e)

    def _get_macros_range(self, document: Document, xml_tree: etree.ElementTree) -> Optional[Range]:
        """Given a XML document and its corresponding ElementTree, finds
        the first macro import element and returns its Range position
        inside the document.

        Args:
            document (Document): The XML tool document.
            xml_tree (etree.ElementTree): The corresponding ElementTree of the document.

        Returns:
            Optional[Range]: The Range position of the import file if it exists.
        """
        try:
            import_element = xml_tree.find("macros//import")
            line_number = import_element.sourceline - 1
            filename = import_element.text
            start = document.lines[line_number].find(filename)
            end = start + len(filename)
            return Range(Position(line_number, start), Position(line_number, end),)
        except BaseException:
            return None

    def _is_macro_definition_file(self, document: Document) -> bool:
        """Determines if a XML document is a macro definition file.

        Args:
            document (Document): The XML document.

        Returns:
            bool: True if it is a macro definition file or False otherwise.
        """
        return document.source.lstrip().startswith(MACROS_TAG)

    def _check_syntax(self, document: Document) -> List[Diagnostic]:
        """Check if the XML document contains any syntax error and returns it in a list.

        Args:
            document (Document): The XML document.

        Returns:
            List[Diagnostic]: The list containing the syntax error found or an empty list.
        """
        try:
            etree.fromstring(document.source)
            return []
        except etree.XMLSyntaxError as e:
            return self._build_diagnostics_from_syntax_error(e)

    def _validate_tree_with_macros(self, document: Document, xml_tree: etree.ElementTree) -> List[Diagnostic]:
        """Validates the document after loading all the macros referenced and expands them.

        Args:
            document (Document): [description]
            xml_tree (etree.ElementTree): [description]

        Returns:
            List[Diagnostic]: [description]
        """
        error_range = None
        try:
            error_range = self._get_macros_range(document, xml_tree)
            expanded_tool_tree, _ = xml_macros.load_with_references(document.path)
            expanded_xml = self._remove_macros(expanded_tool_tree)
            root = expanded_xml.getroot()
            self.xsd_schema.assertValid(root)
            return []
        except etree.DocumentInvalid as e:
            diagnostics = [
                Diagnostic(error_range, f"Validation error on macro: {error.message}", source=self.server_name,)
                for error in e.error_log.filter_from_errors()
            ]
            return diagnostics

        except etree.XMLSyntaxError as e:
            result = Diagnostic(error_range, f"Syntax error on macro: {e.msg}", source=self.server_name,)
            return [result]

    def _validate_tree(self, xml_tree: etree.ElementTree) -> List[Diagnostic]:
        """Validates an XML tree against the XSD schema.

        Args:
            xml_tree (etree.ElementTree): The root element

        Returns:
            List[Diagnostic]: The diagnosis results
        """
        try:
            self.xsd_schema.assertValid(xml_tree)
            return []
        except etree.DocumentInvalid as e:
            return self._build_diagnostics(e.error_log, xml_tree)

    def _remove_macros(self, tool_xml: etree.ElementTree) -> etree.ElementTree:
        """Removes the macros section from the tool tree.

        Args:
            tool_xml (etree.ElementTree): The tool element tree.

        Returns:
            etree.ElementTree: The tool element tree without the macros section.
        """
        to_remove = []
        for macros_el in tool_xml.getroot().findall("macros"):
            to_remove.append(macros_el)
        for macros_el in to_remove:
            tool_xml.getroot().remove(macros_el)
        return tool_xml

    def _build_diagnostics(
        self, error_log: etree._ListErrorLog, xml_tree: Optional[etree.ElementTree] = None
    ) -> List[Diagnostic]:
        """Gets a list of diagnostics from the XSD validation error log.

        Args:
            error_log (etree._ListErrorLog): The error log generated after XSD schema validation.
            xml_tree (Optional[etree.ElementTree], optional): The element tree associated with
            the error log. Defaults to None.

        Raises:
            ExpandMacrosFoundException: This is raised when a macro ``expand`` element is found in the error log.
            The ``expand`` element is not part of the XSD so, when this exception is raised, we know that
            the document must expand the macros before validation.

        Returns:
            List[Diagnostic]: The list of resulting diagnostic items found in the error log.
        """
        diagnostics = []
        for error in error_log.filter_from_errors():
            if "expand" in error.message:
                raise ExpandMacrosFoundException(xml_tree)

            result = Diagnostic(
                Range(Position(error.line - 1, error.column), Position(error.line - 1, error.column),),
                error.message,
                source=self.server_name,
            )
            diagnostics.append(result)
        return diagnostics

    def _build_diagnostics_from_syntax_error(self, error: etree.XMLSyntaxError) -> Diagnostic:
        """Builds a Diagnostic element from a XMLSyntaxError.

        Args:
            error (etree.XMLSyntaxError): The syntax error.

        Returns:
            Diagnostic: The converted Diagnostic item.
        """
        result = Diagnostic(
            Range(Position(error.lineno - 1, error.position[0] - 1), Position(error.lineno - 1, error.position[1] - 1),),
            error.msg,
            source=self.server_name,
        )
        return [result]


class ExpandMacrosFoundException(Exception):
    """This exceptions indicates that the current tool contains
    macros and they should be expanded before validation.
    """

    def __init__(self, xml_tree: Optional[etree.ElementTree] = None):
        self.xml_tree = xml_tree



"""This module contains shared types across all modules.
"""

from pygls.types import Range


class AutoCloseTagResult:
    """Contains the code snippet and the range in the text document that
    will be returned to the client when a tag auto-close is requested.
    """

    def __init__(self, snippet: str, replace_range: Range = None):
        self.snippet = snippet
        self.range = replace_range



import pytest
from typing import List, Optional
from pytest_mock import MockerFixture
from pygls.workspace import Document, Position

from ...services.context import (
    XmlContextService,
    XmlContext,
    XmlContextParser,
    ContextTokenType,
    XsdTree,
    XsdNode,
    Range,
)


# The content starts at line 1 for convenience
FAKE_CONTENT = """
<tool id="test">
    <description/>
    <test value="0"/>
    <help><![CDATA[
        Sample text
    ]]></help>
    <inputs></inputs>
</tool>'
"""
FAKE_DOC_URI = "file://fake_doc.xml"
FAKE_DOCUMENT = Document(FAKE_DOC_URI, FAKE_CONTENT)


def get_fake_document(content: str) -> Document:
    return Document(FAKE_DOC_URI, content)


def print_context_params(document: Document, position: Position) -> None:
    print(f"Test context at position [line={position.line}, char={position.character}]")
    print(f"Document:\n{document.source}")


def get_mock_xsd_tree(mocker: MockerFixture) -> XsdTree:
    root = XsdNode(name="root", element=mocker.Mock())
    child = XsdNode(name="child", parent=root, element=mocker.Mock())
    XsdNode(name="sibling", parent=root, element=mocker.Mock())
    XsdNode(name="subchild", parent=child, element=mocker.Mock())
    return XsdTree(root)


class TestXmlContextClass:
    def test_init_sets_properties(self) -> None:
        expected_line_content = "test"
        expected_position = Position()

        context = XmlContext(document_line=expected_line_content, position=expected_position)

        assert context.document_line == expected_line_content
        assert context.target_position == expected_position
        assert not context.is_empty

    def test_context_with_unknown_token_type_returns_all_false(self) -> None:
        context = XmlContext(token_type=ContextTokenType.UNKNOWN)

        assert not context.is_tag()
        assert not context.is_attribute_key()
        assert not context.is_attribute_value()

    def test_context_with_tag_token_type_returns_is_tag(self) -> None:
        context = XmlContext(token_type=ContextTokenType.TAG)

        assert context.is_tag()
        assert not context.is_attribute_key()
        assert not context.is_attribute_value()

    def test_context_with_attr_key_token_type_returns_is_attr_key(self) -> None:
        context = XmlContext(token_type=ContextTokenType.ATTRIBUTE_KEY)

        assert not context.is_tag()
        assert context.is_attribute_key()
        assert not context.is_attribute_value()

    def test_context_with_attr_value_token_type_returns_is_attr_value(self) -> None:
        context = XmlContext(token_type=ContextTokenType.ATTRIBUTE_VALUE)

        assert not context.is_tag()
        assert not context.is_attribute_key()
        assert context.is_attribute_value()


class TestXmlContextServiceClass:
    def test_init_sets_properties(self, mocker: MockerFixture) -> None:
        expected = mocker.Mock()

        service = XmlContextService(expected)

        assert service.xsd_tree

    def test_get_xml_context_returns_empty_document_context(self, mocker: MockerFixture) -> None:
        xml_content = ""
        position = Position()
        xsd_tree_mock = mocker.Mock()
        service = XmlContextService(xsd_tree_mock)

        context = service.get_xml_context(Document(FAKE_DOC_URI, xml_content), position)

        assert context.is_empty

    @pytest.mark.parametrize(
        "document, position, expected",
        [
            (get_fake_document("<root><child"), Position(line=0, character=0), "root",),
            (get_fake_document("<root><child"), Position(line=0, character=6), "root",),
            (get_fake_document("<root><child"), Position(line=0, character=7), "child",),
            (get_fake_document("<root><child"), Position(line=0, character=12), "child",),
            (get_fake_document("<root><child "), Position(line=0, character=13), "child",),
            (get_fake_document('<root attr="4"><child '), Position(line=0, character=14), "root",),
            (get_fake_document('<root attr="4"><child '), Position(line=0, character=16), "child",),
            (get_fake_document('<root attr="4">\n<child/><other'), Position(line=1, character=7), "child",),
            (get_fake_document('<root attr="4">\n<child/><other'), Position(line=1, character=8), "root",),
            (get_fake_document('<root attr="4">\n<child/><other'), Position(line=1, character=9), "root",),
            (get_fake_document('<root attr="4">\n<child/><sibling'), Position(line=1, character=9), "sibling",),
            (get_fake_document('<root attr="4">\n    <\n<child'), Position(line=1, character=5), "root",),
        ],
    )
    def test_get_xml_context_returns_context_with_expected_node(
        self, mocker: MockerFixture, document, position, expected
    ) -> None:
        xsd_tree_mock = get_mock_xsd_tree(mocker)
        service = XmlContextService(xsd_tree_mock)
        print(xsd_tree_mock.render())
        print_context_params(document, position)

        context = service.get_xml_context(document, position)

        assert context
        assert context.node
        assert context.node.name == expected


class TestXmlContextParserClass:
    @pytest.mark.parametrize(
        "document, position, expected",
        [
            (FAKE_DOCUMENT, Position(line=1, character=0), "tool"),
            (FAKE_DOCUMENT, Position(line=1, character=1), "tool"),
            (FAKE_DOCUMENT, Position(line=1, character=5), None),
            (FAKE_DOCUMENT, Position(line=1, character=6), "id"),
            (FAKE_DOCUMENT, Position(line=1, character=8), "id"),
            (FAKE_DOCUMENT, Position(line=1, character=9), None),
            (FAKE_DOCUMENT, Position(line=1, character=10), "test"),
            (FAKE_DOCUMENT, Position(line=1, character=14), "test"),
            (FAKE_DOCUMENT, Position(line=1, character=15), None),
            (FAKE_DOCUMENT, Position(line=2, character=5), "description"),
            (FAKE_DOCUMENT, Position(line=2, character=15), "description"),
            (FAKE_DOCUMENT, Position(line=2, character=17), None),
            (FAKE_DOCUMENT, Position(line=3, character=4), None),
            (FAKE_DOCUMENT, Position(line=3, character=5), "test"),
            (FAKE_DOCUMENT, Position(line=3, character=8), "test"),
            (FAKE_DOCUMENT, Position(line=3, character=9), None),
            (FAKE_DOCUMENT, Position(line=3, character=10), "value"),
            (FAKE_DOCUMENT, Position(line=3, character=15), "value"),
            (FAKE_DOCUMENT, Position(line=3, character=16), None),
            (FAKE_DOCUMENT, Position(line=3, character=17), "0"),
            (FAKE_DOCUMENT, Position(line=3, character=18), "0"),
            (FAKE_DOCUMENT, Position(line=3, character=19), None),
        ],
    )
    def test_parse_well_formed_xml_return_expected_context_token_name(
        self, document: Document, position: Position, expected: Optional[str]
    ) -> None:
        print_context_params(document, position)
        parser = XmlContextParser()

        context = parser.parse(document, position)

        assert context.token_name == expected

    @pytest.mark.parametrize(
        "document, position, expected",
        [
            (get_fake_document('<other id="1"'), Position(line=0, character=0), None),
            (get_fake_document('<other id="1"'), Position(line=0, character=1), "other"),
            (get_fake_document('<other id="1"'), Position(line=0, character=6), "other"),
            (get_fake_document('<other id="1"'), Position(line=0, character=7), "id"),
            (get_fake_document('<other id="1"'), Position(line=0, character=9), "id"),
            (get_fake_document('<other id="1"'), Position(line=0, character=11), "1"),
            (get_fake_document('<other id="1"'), Position(line=0, character=12), "1"),
            (get_fake_document('<other id="1"'), Position(line=0, character=13), None),
            (get_fake_document('<first id="1" test="value">\n    <second'), Position(line=0, character=2), "first",),
            (get_fake_document('<first id="1" test="value">\n    <second'), Position(line=0, character=8), "id",),
            (get_fake_document('<first id="1" test="value">\n    <second'), Position(line=0, character=11), "1",),
            (get_fake_document('<first id="1" test="value">\n    <second'), Position(line=0, character=12), "1",),
            (get_fake_document('<first id="1" test="value">\n    <second'), Position(line=0, character=14), "test",),
            (get_fake_document('<first id="1" test="value">\n    <second'), Position(line=0, character=20), "value",),
            (get_fake_document('<first id="1" test="value">\n    <second'), Position(line=0, character=25), "value",),
            (get_fake_document('<first id="1" test="value">\n    <second'), Position(line=1, character=5), "second",),
            (get_fake_document('<first id="one_test">\n    <second'), Position(line=0, character=11), "one_test",),
            (get_fake_document('<first id="one_test">\n    <second'), Position(line=0, character=19), "one_test",),
            (get_fake_document('<first id="one test">\n    <second'), Position(line=0, character=11), "one test",),
            (get_fake_document('<first id="one test">\n    <second'), Position(line=0, character=19), "one test",),
            (get_fake_document('<first id="one test'), Position(line=0, character=19), "one test",),
            (get_fake_document('<first id="value"><second/></first>'), Position(line=0, character=19), "second",),
            (get_fake_document('<first id="value"><second/></first>'), Position(line=0, character=29), "first",),
        ],
    )
    def test_parse_incomplete_xml_return_expected_context_token_name(
        self, document: Document, position: Position, expected: str
    ) -> None:
        print_context_params(document, position)
        parser = XmlContextParser()

        context = parser.parse(document, position)

        assert context.token_name == expected

    @pytest.mark.parametrize(
        "document, position, expected",
        [
            (
                get_fake_document('<first id="1" test="value">\n    <second'),
                Position(line=0, character=0),
                ContextTokenType.TAG,
            ),
            (
                get_fake_document('<first id="1" test="value">\n    <second'),
                Position(line=0, character=2),
                ContextTokenType.TAG,
            ),
            (
                get_fake_document('<first id="1" test="value">\n    <second'),
                Position(line=0, character=8),
                ContextTokenType.ATTRIBUTE_KEY,
            ),
            (
                get_fake_document('<first id="1" test="value">\n    <second'),
                Position(line=0, character=11),
                ContextTokenType.ATTRIBUTE_VALUE,
            ),
            (
                get_fake_document('<first id="one_test">\n    <second'),
                Position(line=0, character=19),
                ContextTokenType.ATTRIBUTE_VALUE,
            ),
            (
                get_fake_document('<first id="1" test="value">\n    <second'),
                Position(line=1, character=0),
                ContextTokenType.UNKNOWN,
            ),
            (
                get_fake_document('<first id="1" test="value">\n    <second'),
                Position(line=1, character=4),
                ContextTokenType.UNKNOWN,
            ),
            (get_fake_document('<first id="one test'), Position(line=0, character=19), ContextTokenType.ATTRIBUTE_VALUE,),
            (get_fake_document('<first id="1">value</first>'), Position(line=0, character=14), ContextTokenType.UNKNOWN,),
            (get_fake_document('<first id="value"><second/></first>'), Position(line=0, character=19), ContextTokenType.TAG,),
            (get_fake_document('<first id="value"><second/></first>'), Position(line=0, character=29), ContextTokenType.TAG,),
            (get_fake_document('<first id="value"><second/></first'), Position(line=0, character=29), ContextTokenType.TAG,),
        ],
    )
    def test_parse_incomplete_xml_return_expected_context_token_type(
        self, document: Document, position: Position, expected: ContextTokenType
    ) -> None:
        print_context_params(document, position)
        parser = XmlContextParser()

        context = parser.parse(document, position)

        assert context.token_type == expected

    @pytest.mark.parametrize(
        "document, position, expected",
        [
            (get_fake_document(""), Position(line=0, character=0), True),
            (get_fake_document(""), Position(line=0, character=10), True),
            (get_fake_document("<"), Position(line=0, character=0), True),
            (get_fake_document("   "), Position(line=0, character=0), True),
            (get_fake_document("\n\n"), Position(line=0, character=0), True),
            (get_fake_document("\n\n"), Position(line=1, character=0), True),
        ],
    )
    def test_parse_empty_xml_return_empty_context(self, document: Document, position: Position, expected: bool) -> None:
        print_context_params(document, position)
        parser = XmlContextParser()

        context = parser.parse(document, position)

        assert context.is_empty == expected

    @pytest.mark.parametrize(
        "document, position, expected",
        [
            (get_fake_document("<first><second><third>"), Position(line=0, character=1), ["first"],),
            (get_fake_document("<first><second><third>"), Position(line=0, character=8), ["first", "second"],),
            (get_fake_document("<first><second><third>"), Position(line=0, character=16), ["first", "second", "third"],),
            (get_fake_document("<first><second/><third>"), Position(line=0, character=8), ["first", "second"],),
            (get_fake_document("<first><second/><third>"), Position(line=0, character=17), ["first", "third"],),
            (get_fake_document("<first><second></second><third"), Position(line=0, character=25), ["first", "third"],),
            (get_fake_document("<first><second></second><third"), Position(line=0, character=18), ["first", "second"],),
            (get_fake_document("<first><second\n<third/></first>"), Position(line=1, character=10), ["first"],),
            (get_fake_document("<first><second/></first>"), Position(line=0, character=18), ["first"],),
            (get_fake_document("<first><second/>\n</first>"), Position(line=1, character=2), ["first"],),
            (get_fake_document('<first><second attr="value"/><third'), Position(line=0, character=29), ["first"],),
            (get_fake_document('<first><second attr="value"/><third'), Position(line=0, character=30), ["first", "third"],),
            (get_fake_document('<first attr="value"><third>\n'), Position(line=0, character=26), ["first", "third"],),
            (get_fake_document("<first>\n    <second>  \n</first>"), Position(line=1, character=11), ["first", "second"],),
            (FAKE_DOCUMENT, Position(line=4, character=10), ["tool", "help"]),
            (FAKE_DOCUMENT, Position(line=4, character=18), ["tool", "help"]),
            (FAKE_DOCUMENT, Position(line=6, character=5), ["tool", "help"]),
            (FAKE_DOCUMENT, Position(line=6, character=7), ["tool", "help"]),
            (
                get_fake_document('<first>\n<second attr="value"\n</second>'),
                Position(line=2, character=2),
                ["first", "second"],
            ),
        ],
    )
    def test_parse_return_expected_tag_stack_context(
        self, document: Document, position: Position, expected: List[str]
    ) -> None:
        print_context_params(document, position)
        parser = XmlContextParser()

        context = parser.parse(document, position)

        assert context.node_stack == expected

    @pytest.mark.parametrize(
        "document, position, expected",
        [
            (
                get_fake_document("<first><second><third>"),
                Position(line=0, character=1),
                Range(start=Position(line=0, character=1), end=Position(line=0, character=6)),
            ),
            (
                get_fake_document("<first><second><third>"),
                Position(line=0, character=8),
                Range(start=Position(line=0, character=8), end=Position(line=0, character=14)),
            ),
            (
                get_fake_document("<first><second/><third>"),
                Position(line=0, character=8),
                Range(start=Position(line=0, character=8), end=Position(line=0, character=14)),
            ),
            (
                get_fake_document("<first><second/><third>"),
                Position(line=0, character=17),
                Range(start=Position(line=0, character=17), end=Position(line=0, character=22)),
            ),
            (
                get_fake_document("<first><second></second><third>"),
                Position(line=0, character=25),
                Range(start=Position(line=0, character=25), end=Position(line=0, character=30)),
            ),
            (
                get_fake_document("<first><second/>\n</first>"),
                Position(line=1, character=2),
                Range(start=Position(line=1, character=0), end=Position(line=1, character=8)),
            ),
            (
                get_fake_document('<first attr="value"/><third'),
                Position(line=0, character=8),
                Range(start=Position(line=0, character=7), end=Position(line=0, character=11)),
            ),
            (
                get_fake_document('<first>\n<second attr="value"/>\n<third'),
                Position(line=1, character=15),
                Range(start=Position(line=1, character=14), end=Position(line=1, character=19)),
            ),
            (
                get_fake_document('<first attr1="value1" attr2="value2">'),
                Position(line=0, character=24),
                Range(start=Position(line=0, character=22), end=Position(line=0, character=27)),
            ),
            (
                get_fake_document('<first attr1="value1" attr2="value2">'),
                Position(line=0, character=30),
                Range(start=Position(line=0, character=29), end=Position(line=0, character=35)),
            ),
        ],
    )
    def test_parse_return_expected_tag_range_context(self, document: Document, position: Position, expected: Range) -> None:
        print_context_params(document, position)
        parser = XmlContextParser()

        context = parser.parse(document, position)

        assert context.token_range == expected

    @pytest.mark.parametrize(
        "document, position, expected",
        [
            (
                get_fake_document('<first id="1">\n    <second attr="value">test</second>\n<third'),
                Position(line=0, character=1),
                False,
            ),
            (
                get_fake_document('<first id="1">\n    <second attr="value">test</second>\n<third'),
                Position(line=0, character=8),
                False,
            ),
            (
                get_fake_document('<first id="1">\n    <second attr="value">test</second>\n<third'),
                Position(line=0, character=14),
                True,
            ),
            (
                get_fake_document('<first id="1">\n    <second attr="value">test</second>\n<third'),
                Position(line=1, character=0),
                True,
            ),
            (
                get_fake_document('<first id="1">\n    <second attr="value">test</second>\n<third'),
                Position(line=1, character=4),
                True,
            ),
            (
                get_fake_document('<first id="1">\n    <second attr="value">test</second>\n<third'),
                Position(line=1, character=5),
                False,
            ),
            (
                get_fake_document('<first id="1">\n    <second attr="value">test</second>\n<third'),
                Position(line=1, character=6),
                False,
            ),
            (
                get_fake_document('<first id="1">\n    <second attr="value">test</second>\n<third'),
                Position(line=1, character=24),
                False,
            ),
            (
                get_fake_document('<first id="1">\n    <second attr="value">test</second>\n<third'),
                Position(line=1, character=25),
                True,
            ),
            (
                get_fake_document('<first id="1">\n    <second attr="value">test</second>\n<third'),
                Position(line=1, character=29),
                True,
            ),
            (
                get_fake_document('<first id="1">\n    <second attr="value">test</second>\n<third'),
                Position(line=1, character=30),
                False,
            ),
            (FAKE_DOCUMENT, Position(line=4, character=10), True),
            (FAKE_DOCUMENT, Position(line=4, character=18), True),
            (FAKE_DOCUMENT, Position(line=6, character=5), True),
            (FAKE_DOCUMENT, Position(line=6, character=7), True),
        ],
    )
    def test_parse_returns_context_with_expected_is_node_content(
        self, document: Document, position: Position, expected: bool
    ) -> None:
        print_context_params(document, position)
        parser = XmlContextParser()

        context = parser.parse(document, position)

        assert context.is_node_content == expected

    @pytest.mark.parametrize(
        "document, position, expected",
        [
            (get_fake_document('<first id="1" test="2"'), Position(line=0, character=2), ["id", "test"],),
            (get_fake_document('<first id="1" test="2"'), Position(line=0, character=14), ["id", "test"],),
            (get_fake_document('<first id="1" test="2"'), Position(line=0, character=22), ["id", "test"],),
            (get_fake_document('<first id="1" test="2"/>'), Position(line=0, character=2), ["id", "test"],),
            (get_fake_document('<first id="1" test="2">\n</first>'), Position(line=0, character=10), ["id", "test"],),
        ],
    )
    def test_parse_return_expected_attributes_at_node(
        self, document: Document, position: Position, expected: List[str]
    ) -> None:
        print_context_params(document, position)
        parser = XmlContextParser()

        context = parser.parse(document, position)

        assert context.attr_list == expected

    @pytest.mark.parametrize(
        "document, position, expected",
        [
            (get_fake_document("<first> </first>"), Position(line=0, character=7), False,),
            (get_fake_document("<first></first>"), Position(line=0, character=8), True,),
            (get_fake_document("<first></first>"), Position(line=0, character=14), True,),
            (get_fake_document("<first></first >"), Position(line=0, character=14), True,),
            (get_fake_document("<first></first"), Position(line=0, character=10), True,),
            (get_fake_document("<first/>"), Position(line=0, character=7), True,),
            (get_fake_document("<first/ >"), Position(line=0, character=7), True,),
            (get_fake_document('<first attr="1"/>'), Position(line=0, character=15), False,),
            (get_fake_document('<first attr="1" />'), Position(line=0, character=15), False,),
            (get_fake_document('<first attr="1"/'), Position(line=0, character=15), False,),
        ],
    )
    def test_parse_return_expected_is_closin_tag(self, document: Document, position: Position, expected: bool) -> None:
        print_context_params(document, position)
        parser = XmlContextParser()

        context = parser.parse(document, position)

        assert context.is_closing_tag == expected



"""Utilities to validate Galaxy xml tool wrappers and extract
information from the XSD schema.
"""

from typing import List
from lxml import etree

from pygls.workspace import Document
from pygls.types import Diagnostic, MarkupContent, MarkupKind

from .constants import TOOL_XSD_FILE, MSG_NO_DOCUMENTATION_AVAILABLE
from .parser import GalaxyToolXsdParser
from .validation import GalaxyToolValidationService
from ..context import XmlContext

NO_DOC_MARKUP = MarkupContent(MarkupKind.Markdown, MSG_NO_DOCUMENTATION_AVAILABLE)


class GalaxyToolXsdService:
    """Galaxy tool Xml Schema Definition service.

    This service provides functionality to extract information from
    the XSD schema and validate XML files against it.
    """

    def __init__(self, server_name: str):
        """Initializes the validator by loading the XSD."""
        self.server_name = server_name
        self.xsd_doc = etree.parse(str(TOOL_XSD_FILE))
        self.xsd_schema = etree.XMLSchema(self.xsd_doc)
        self.xsd_parser = GalaxyToolXsdParser(self.xsd_doc.getroot())
        self.validator = GalaxyToolValidationService(server_name, self.xsd_schema)

    def validate_document(self, document: Document) -> List[Diagnostic]:
        """Validates the Galaxy tool xml using the XSD schema and returns a list
        of diagnotics if there are any problems.
        """
        return self.validator.validate_document(document)

    def get_documentation_for(self, context: XmlContext) -> MarkupContent:
        """Gets the documentation annotated in the XSD about the
        given element name (node or attribute).
        """
        tree = self.xsd_parser.get_tree()
        node = tree.find_node_by_stack(context.node_stack)
        if node is None:
            return NO_DOC_MARKUP
        element = None
        if context.is_tag():
            element = node
        if context.is_attribute_key():
            element = node.attributes.get(context.token_name)
        if element is None:
            return NO_DOC_MARKUP
        return element.get_doc()



from .xsd.service import GalaxyToolXsdService
from .format import GalaxyToolFormatService
from .completion import XmlCompletionService, AutoCloseTagResult
from .context import XmlContextService

from typing import List, Optional
from pygls.workspace import Document
from pygls.types import (
    CompletionList,
    CompletionParams,
    Diagnostic,
    DocumentFormattingParams,
    Hover,
    Position,
    TextDocumentPositionParams,
    TextEdit,
)


class GalaxyToolLanguageService:
    """Galaxy tool Language service.

    This service manages all the operations supported
    by the LSP.
    """

    def __init__(self, server_name: str):
        self.xsd_service = GalaxyToolXsdService(server_name)
        self.format_service = GalaxyToolFormatService()
        tree = self.xsd_service.xsd_parser.get_tree()
        self.completion_service = XmlCompletionService(tree)
        self.xml_context_service = XmlContextService(tree)

    def get_diagnostics(self, document: Document) -> List[Diagnostic]:
        """Validates the Galaxy tool and returns a list
        of diagnotics if there are any problems.
        """
        return self.xsd_service.validate_document(document)

    def get_documentation(self, document: Document, position: Position) -> Optional[Hover]:
        """Gets the documentation about the element at the given position."""
        context = self.xml_context_service.get_xml_context(document, position)
        if context.is_node_content or context.is_attribute_value():
            return None
        documentation = self.xsd_service.get_documentation_for(context)
        return Hover(documentation, context.token_range)

    def format_document(self, content: str, params: DocumentFormattingParams) -> List[TextEdit]:
        """Given the document contents returns the list of TextEdits
        needed to properly format and layout the document.
        """
        return self.format_service.format(content, params)

    def get_completion(self, document: Document, params: CompletionParams) -> CompletionList:
        """Gets completion items depending on the current document context."""
        context = self.xml_context_service.get_xml_context(document, params.position)
        return self.completion_service.get_completion_at_context(context, params.context)

    def get_auto_close_tag(self, document: Document, params: TextDocumentPositionParams) -> AutoCloseTagResult:
        """Gets the closing result for the currently opened tag in context."""
        trigger_character = document.lines[params.position.line][params.position.character - 1]
        position_before_trigger = Position(params.position.line, params.position.character - 1)
        context = self.xml_context_service.get_xml_context(document, position_before_trigger)
        return self.completion_service.get_auto_close_tag(context, trigger_character)



"""Module in charge of the auto-completion feature."""

from pygls.types import (
    CompletionContext,
    CompletionItem,
    CompletionItemKind,
    CompletionList,
    CompletionTriggerKind,
    InsertTextFormat,
    Position,
    Range,
)

from .xsd.parser import XsdTree, XsdNode, XsdAttribute
from .context import XmlContext
from ..types import AutoCloseTagResult


class XmlCompletionService:
    """Service in charge of generating completion lists based
    on the current XML context.
    """

    def __init__(self, xsd_tree: XsdTree):
        self.xsd_tree: XsdTree = xsd_tree

    def get_completion_at_context(self, xml_context: XmlContext, completion_context: CompletionContext) -> CompletionList:
        triggerKind = completion_context.triggerKind
        if triggerKind == CompletionTriggerKind.TriggerCharacter:
            if completion_context.triggerCharacter == "<":
                return self.get_node_completion(xml_context)
            if completion_context.triggerCharacter == " ":
                return self.get_attribute_completion(xml_context)
        elif triggerKind == CompletionTriggerKind.Invoked:
            if xml_context.is_attribute_value():
                return self.get_attribute_value_completion(xml_context)
        return CompletionList(items=[], is_incomplete=False)

    def get_node_completion(self, context: XmlContext) -> CompletionList:
        """Gets a list of completion items with all the available child tags
        that can be placed in the current context node.

        Args:
            context (XmlContext): The XML context information at a specific
            document position. It should contain, at least, the current node.

        Returns:
            CompletionList: A list of completion items with the child nodes
            that can be placed under the current node.
        """
        result = []
        if context.is_empty:
            result.append(self._build_node_completion_item(context.node))
        elif context.node:
            for child in context.node.children:
                result.append(self._build_node_completion_item(child, len(result)))
            result.append(self._build_node_completion_item(self.xsd_tree.expand_element, len(result)))
        return CompletionList(items=result, is_incomplete=False)

    def get_attribute_completion(self, context: XmlContext) -> CompletionList:
        """Gets a list of completion items with all the attributes that can be
        used in the current context node.

        Args:
            context (XmlContext): The XML context information at a specific
            document position. It should contain, at least, the current node.

        Returns:
            CompletionList: The completion item with the basic information
            about the attributes.
        """
        if context.is_empty or context.is_node_content or context.is_attribute_value() or context.is_closing_tag:
            return CompletionList(is_incomplete=False)

        result = []
        if context.node:
            for attr_name in context.node.attributes:
                if attr_name in context.attr_list:
                    continue
                attr = context.node.attributes[attr_name]
                result.append(self._build_attribute_completion_item(attr, len(result)))
        return CompletionList(items=result, is_incomplete=False)

    def get_attribute_value_completion(self, context: XmlContext) -> CompletionList:
        """Gets a list of possible values for a anumeration restricted attribute if exists.

        Args:
            context (XmlContext): The XML context at an attribute value position.

        Returns:
            CompletionList: The list of possible values of the attribute if it has an enumeration
            restriction.
        """
        attribute: XsdAttribute = context.node.attributes.get(context.attr_name)
        if attribute and attribute.enumeration:
            result = [CompletionItem(item, CompletionItemKind.Value) for item in attribute.enumeration]
            return CompletionList(items=result, is_incomplete=False)

    def get_auto_close_tag(self, context: XmlContext, trigger_character: str) -> AutoCloseTagResult:
        """Gets the closing result for the currently opened tag in context."""
        tag = context.node.name
        snippet = f"$0</{tag}>"
        replace_range = None
        is_self_closing = trigger_character == "/"
        if is_self_closing:
            start = Position(context.target_position.line, context.target_position.character)
            end_character = context.target_position.character + 1
            if len(context.document_line) > end_character and context.document_line[end_character] == ">":
                end_character = end_character + 1
            end = Position(context.target_position.line, end_character)
            replace_range = Range(start=start, end=end)
            if not context.is_node_content:
                snippet = "/>$0"
        elif context.is_node_content:
            return None

        return AutoCloseTagResult(snippet, replace_range)

    def _build_node_completion_item(self, node: XsdNode, order: int = 0) -> CompletionItem:
        """Generates a completion item with the information about the
        given node definition.

        Args:
            node (XsdNode): The node definition used to build the
            completion item.
            order (int): The position for ordering this item.

        Returns:
            CompletionItem: The completion item with the basic information
            about the node.
        """
        return CompletionItem(
            node.name, CompletionItemKind.Class, documentation=node.get_doc(), sort_text=str(order).zfill(2),
        )

    def _build_attribute_completion_item(self, attr: XsdAttribute, order: int = 0) -> CompletionItem:
        """Generates a completion item with the information about the
        given attribute definition.

        Args:
            attr (XsdAttribute): The attribute definition used to build the
            completion item.
            order (int): The position for ordering this item.

        Returns:
            CompletionItem: The completion item with the basic information
            about the attribute.
        """
        return CompletionItem(
            attr.name,
            CompletionItemKind.Variable,
            documentation=attr.get_doc(),
            insert_text=f'{attr.name}="$1"',
            insert_text_format=InsertTextFormat.Snippet,
            sort_text=str(order).zfill(2),
        )



import pytest

from typing import List
from lxml import etree

from .sample_data import TEST_XSD, RECURSIVE_XSD
from ...services.xsd.parser import GalaxyToolXsdParser
from ...services.xsd.constants import MSG_NO_DOCUMENTATION_AVAILABLE


@pytest.fixture(scope="module")
def xsd_parser() -> GalaxyToolXsdParser:
    root = etree.fromstring(TEST_XSD)
    parser = GalaxyToolXsdParser(root)
    return parser


@pytest.fixture(scope="module")
def recursive_xsd_parser() -> GalaxyToolXsdParser:
    root = etree.fromstring(RECURSIVE_XSD)
    parser = GalaxyToolXsdParser(root)
    return parser


class TestXsdParserClass:
    def test_get_tree_returns_valid_element_names(self, xsd_parser: GalaxyToolXsdParser) -> None:
        tree = xsd_parser.get_tree()
        root = tree.root

        # If the test fails the tree will be printed
        print(tree.render())

        assert len(root.ancestors) == 0
        assert len(root.descendants) == 10
        assert root.name == "testElement"
        assert root.children[0].name == "firstElement"
        assert root.children[1].name == "secondElement"
        assert root.children[1].children[0].name == "group_elem1"
        assert root.children[1].children[1].name == "group_elem2"
        assert root.children[1].children[2].name == "content"
        assert root.children[2].name == "thirdElement"
        assert root.children[2].children[0].name == "childElement"
        assert root.children[3].name == "element_with_group"
        assert root.children[3].children[0].name == "group_elem1"
        assert root.children[3].children[1].name == "group_elem2"

    def test_get_tree_returns_valid_attribute_names_using_groups(self, xsd_parser: GalaxyToolXsdParser) -> None:
        tree = xsd_parser.get_tree()
        root = tree.root

        assert len(root.attributes) == 4
        assert root.attributes["id"]
        assert root.attributes["value"]
        assert root.attributes["gattr1"]
        assert root.attributes["gattr2"]

    def test_get_tree_returns_valid_attribute_names_when_simple_content(self, xsd_parser: GalaxyToolXsdParser) -> None:
        tree = xsd_parser.get_tree()
        simple_content_elem = tree.root.children[2].children[0]

        assert len(simple_content_elem.attributes) == 1
        assert simple_content_elem.attributes["simple"]

    def test_get_tree_returns_valid_element_when_complex_content(self, xsd_parser: GalaxyToolXsdParser) -> None:
        tree = xsd_parser.get_tree()
        complex_content_elem = tree.root.children[2].children[0]

        assert len(complex_content_elem.attributes) == 1
        assert complex_content_elem.attributes["simple"]

    def test_get_tree_with_recursive_xsd_stops_recursion(self, recursive_xsd_parser: GalaxyToolXsdParser) -> None:
        tree = recursive_xsd_parser.get_tree()

        assert len(tree.root.descendants) > 0

    def test_tree_find_node_by_name_returns_expected_node(self, xsd_parser: GalaxyToolXsdParser) -> None:
        tree = xsd_parser.get_tree()
        expected = "childElement"

        node = tree.find_node_by_name(expected)

        assert node.name == expected

    @pytest.mark.parametrize(
        "stack, expected",
        [
            (["testElement"], "testElement",),
            (["testElement", "firstElement"], "firstElement",),
            (["testElement", "secondElement"], "secondElement",),
            (["testElement", "secondElement", "group_elem1"], "group_elem1",),
            (["testElement", "element_with_group", "group_elem1"], "group_elem1",),
        ],
    )
    def test_tree_find_node_by_stack_returns_expected_node(
        self, xsd_parser: GalaxyToolXsdParser, stack: List[str], expected: str
    ) -> None:
        tree = xsd_parser.get_tree()

        node = tree.find_node_by_stack(stack)

        assert node.name == expected

    @pytest.mark.parametrize(
        "stack, expected",
        [
            (["expand"], "expand",),
            (["testElement", "expand"], "expand",),
            (["testElement", "element_with_group", "expand"], "expand",),
        ],
    )
    def test_tree_find_node_by_stack_ending_with_expand_returns_expand_node(
        self, xsd_parser: GalaxyToolXsdParser, stack: List[str], expected: str
    ) -> None:
        tree = xsd_parser.get_tree()

        node = tree.find_node_by_stack(stack)

        assert node.name == expected

    def test_tree_find_node_by_name_returns_None_when_node_not_found(self, xsd_parser: GalaxyToolXsdParser) -> None:
        tree = xsd_parser.get_tree()

        node = tree.find_node_by_name("unknown")

        assert node is None

    def test_get_documentation_returns_valid_when_exists(self, xsd_parser: GalaxyToolXsdParser) -> None:
        tree = xsd_parser.get_tree()

        doc = tree.root.get_doc()

        assert doc.value == "Documentation ``example``."

    def test_get_documentation_returns_valid_other_language(self, xsd_parser: GalaxyToolXsdParser) -> None:
        tree = xsd_parser.get_tree()

        doc = tree.root.get_doc("es")

        assert doc.value == "``Ejemplo`` de documentación."

    def test_get_documentation_should_return_no_documentation_when_not_exists(self, xsd_parser: GalaxyToolXsdParser) -> None:
        tree = xsd_parser.get_tree()

        doc = tree.root.get_doc("de")

        assert doc.value == MSG_NO_DOCUMENTATION_AVAILABLE

    def test_parser_returns_expected_enumeration_restrictions(self, xsd_parser: GalaxyToolXsdParser) -> None:
        tree = xsd_parser.get_tree()

        attribute_with_restriction = tree.root.attributes["value"]
        actual = attribute_with_restriction.enumeration

        assert actual == ["v1", "v2", "v3"]

    def test_parser_returns_expected_occurs_restrictions(self, xsd_parser: GalaxyToolXsdParser) -> None:
        tree = xsd_parser.get_tree()

        actual = tree.root.children[2].children[0]

        assert actual.min_occurs == 1
        assert actual.max_occurs == 1

    def test_parser_returns_expected_occurs_when_unbounded(self, xsd_parser: GalaxyToolXsdParser) -> None:
        tree = xsd_parser.get_tree()

        actual = tree.root.children[0]

        assert actual.min_occurs == 0
        assert actual.max_occurs == -1



import pytest
from pytest_mock import MockerFixture
from ...services.completion import (
    AutoCloseTagResult,
    CompletionContext,
    CompletionTriggerKind,
    CompletionItemKind,
    Position,
    Range,
    XmlCompletionService,
    XmlContext,
    XsdTree,
    XsdNode,
    XsdAttribute,
)
from ...services.context import ContextTokenType


@pytest.fixture()
def fake_tree(mocker: MockerFixture) -> XsdTree:
    fake_root = XsdNode("root", element=mocker.Mock())
    fake_attr = XsdAttribute("attr", element=mocker.Mock())
    fake_attr.enumeration = ["v1", "v2"]
    fake_root.attributes[fake_attr.name] = fake_attr
    XsdNode("child", element=mocker.Mock(), parent=fake_root)
    return XsdTree(fake_root)


@pytest.fixture()
def fake_tree_with_attrs(mocker: MockerFixture) -> XsdTree:
    fake_root = XsdNode("root", element=mocker.Mock())
    fake_attr = XsdAttribute("one", element=mocker.Mock())
    fake_root.attributes[fake_attr.name] = fake_attr
    fake_attr = XsdAttribute("two", element=mocker.Mock())
    fake_root.attributes[fake_attr.name] = fake_attr
    fake_attr = XsdAttribute("three", element=mocker.Mock())
    fake_root.attributes[fake_attr.name] = fake_attr
    XsdNode("child", element=mocker.Mock(), parent=fake_root)
    return XsdTree(fake_root)


@pytest.fixture()
def fake_empty_context(fake_tree: XsdTree) -> XmlContext:
    fake_root = fake_tree.root
    fake_context = XmlContext(is_empty=True)
    fake_context.node = fake_root
    return fake_context


@pytest.fixture()
def fake_context_on_root_node(fake_tree: XsdTree) -> XmlContext:
    fake_root = fake_tree.root
    fake_context = XmlContext()
    fake_context.is_empty = False
    fake_context.token_type = ContextTokenType.TAG
    fake_context.token_name = fake_root.name
    fake_context.node = fake_root
    return fake_context


def get_fake_context_with_line_position(fake_tree: XsdTree, line: str, position: Position) -> XmlContext:
    fake_root = fake_tree.root
    fake_context = XmlContext()
    fake_context.document_line = line
    fake_context.target_position = position
    fake_context.node = fake_root
    return fake_context


class TestXmlCompletionServiceClass:
    def test_init_sets_properties(self, fake_tree: XsdTree) -> None:

        service = XmlCompletionService(fake_tree)

        assert service.xsd_tree

    def test_get_completion_at_context_with_open_tag_trigger_returns_expected_node(self, fake_tree: XsdTree) -> None:
        fake_context = XmlContext()
        fake_context.is_empty = False
        fake_context.token_type = ContextTokenType.UNKNOWN
        fake_context.node = fake_tree.root
        fake_completion_context = CompletionContext(CompletionTriggerKind.TriggerCharacter, trigger_character="<")
        service = XmlCompletionService(fake_tree)

        actual = service.get_completion_at_context(fake_context, fake_completion_context)

        assert actual
        assert len(actual.items) == 2
        assert actual.items[0].label == "child"
        assert actual.items[0].kind == CompletionItemKind.Class
        assert actual.items[1].label == "expand"
        assert actual.items[1].kind == CompletionItemKind.Class

    def test_get_completion_at_context_on_node_returns_expected_attributes(self, fake_tree: XsdTree) -> None:
        fake_context = XmlContext()
        fake_context.is_empty = False
        fake_context.token_type = ContextTokenType.UNKNOWN
        fake_context.node = fake_tree.root
        fake_completion_context = CompletionContext(CompletionTriggerKind.TriggerCharacter, trigger_character=" ")
        service = XmlCompletionService(fake_tree)

        actual = service.get_completion_at_context(fake_context, fake_completion_context)

        assert actual
        assert len(actual.items) == 1
        assert actual.items[0].label == "attr"
        assert actual.items[0].kind == CompletionItemKind.Variable

    def test_get_completion_at_context_on_node_with_attribute_returns_expected_attributes(
        self, fake_tree_with_attrs: XsdTree
    ) -> None:
        fake_context = XmlContext()
        fake_context.is_empty = False
        fake_context.attr_list = ["one"]
        fake_context.token_type = ContextTokenType.UNKNOWN
        fake_context.node = fake_tree_with_attrs.root
        fake_completion_context = CompletionContext(CompletionTriggerKind.TriggerCharacter, trigger_character=" ")
        service = XmlCompletionService(fake_tree_with_attrs)

        actual = service.get_completion_at_context(fake_context, fake_completion_context)

        assert actual
        assert len(actual.items) == 2
        assert actual.items[0].label == "two"
        assert actual.items[0].kind == CompletionItemKind.Variable
        assert actual.items[1].label == "three"
        assert actual.items[1].kind == CompletionItemKind.Variable

    def test_get_completion_at_context_on_attr_value_returns_expected_enums(self, fake_tree: XsdTree) -> None:
        fake_context = XmlContext()
        fake_context.is_empty = False
        fake_context.token_type = ContextTokenType.ATTRIBUTE_VALUE
        fake_context.attr_name = "attr"
        fake_context.node = fake_tree.root
        fake_completion_context = CompletionContext(CompletionTriggerKind.Invoked)
        service = XmlCompletionService(fake_tree)

        actual = service.get_completion_at_context(fake_context, fake_completion_context)

        assert actual
        assert len(actual.items) == 2
        assert actual.items[0].label == "v1"
        assert actual.items[0].kind == CompletionItemKind.Value
        assert actual.items[1].label == "v2"
        assert actual.items[1].kind == CompletionItemKind.Value

    def test_return_valid_completion_with_node_context(self, fake_tree: XsdTree, fake_context_on_root_node) -> None:
        service = XmlCompletionService(fake_tree)

        actual = service.get_node_completion(fake_context_on_root_node)

        assert len(actual.items) == 2
        assert actual.items[0].label == fake_tree.root.children[0].name
        assert actual.items[1].label == "expand"

    def test_completion_return_root_node_when_empty_context(self, fake_tree: XsdTree, fake_empty_context) -> None:
        service = XmlCompletionService(fake_tree)

        actual = service.get_node_completion(fake_empty_context)

        assert len(actual.items) == 1
        assert actual.items[0].label == fake_tree.root.name

    def test_return_empty_attribute_completion_when_empty_context(self, fake_tree: XsdTree, fake_empty_context) -> None:
        service = XmlCompletionService(fake_tree)

        actual = service.get_attribute_completion(fake_empty_context)

        assert len(actual.items) == 0

    def test_return_valid_attribute_completion_when_node_context(self, fake_tree: XsdTree, fake_context_on_root_node) -> None:
        service = XmlCompletionService(fake_tree)

        actual = service.get_attribute_completion(fake_context_on_root_node)

        assert len(actual.items) > 0

    def test_return_valid_attribute_value_completion_when_enum_context(self, fake_tree: XsdTree) -> None:
        fake_context = XmlContext()
        fake_context.is_empty = False
        fake_context.token_type = ContextTokenType.ATTRIBUTE_VALUE
        fake_context.attr_name = "attr"
        fake_context.node = fake_tree.root

        service = XmlCompletionService(fake_tree)

        actual = service.get_attribute_value_completion(fake_context)

        assert len(actual.items) == 2

    @pytest.mark.parametrize(
        "context_line, position, trigger, expected",
        [("<root>", Position(line=0, character=6), ">", "$0</root>"), ("<root/", Position(line=0, character=0), "/", "/>$0")],
    )
    def test_auto_close_returns_expected_snippet_at_context(
        self, fake_tree: XsdTree, context_line: str, position: Position, trigger: str, expected: str,
    ) -> None:
        service = XmlCompletionService(fake_tree)
        fake_context = get_fake_context_with_line_position(fake_tree, context_line, position)

        actual = service.get_auto_close_tag(fake_context, trigger)

        assert actual.snippet == expected

    @pytest.mark.parametrize(
        "context_line, position, trigger, expected",
        [
            ("<root>", Position(line=0, character=6), ">", None),
            ("<root/", Position(line=0, character=5), "/", Range(Position(character=5), Position(character=6)),),
            ("<root/>", Position(line=0, character=5), "/", Range(Position(character=5), Position(character=7)),),
        ],
    )
    def test_auto_close_returns_expected_replace_range_at_context(
        self, fake_tree: XsdTree, context_line: str, position: Position, trigger: str, expected: str,
    ) -> None:
        service = XmlCompletionService(fake_tree)
        fake_context = get_fake_context_with_line_position(fake_tree, context_line, position)

        actual = service.get_auto_close_tag(fake_context, trigger)

        assert actual.range == expected

    def test_auto_close_returns_none_when_at_node_content(
        self, fake_tree: XsdTree, fake_context_on_root_node: XmlContext
    ) -> None:
        service = XmlCompletionService(fake_tree)
        trigger = ">"
        fake_context = fake_context_on_root_node
        fake_context.is_node_content = True

        actual = service.get_auto_close_tag(fake_context, trigger)

        assert not actual


class TestAutoCloseTagResultClass:
    def test_init_sets_properties(self, mocker: MockerFixture) -> None:
        expected_snippet = "snippet"
        expected_replace_range = mocker.Mock()

        result = AutoCloseTagResult(expected_snippet, expected_replace_range)

        assert result.snippet == expected_snippet
        assert result.range == expected_replace_range



from ...services.format import GalaxyToolFormatService
from pygls.types import (
    DocumentFormattingParams,
    TextDocumentIdentifier,
    FormattingOptions,
)

FAKE_INVALID_DOCUMENT = """
<invalid> XML content
"""

FAKE_UNFORMATTED_DOCUMENT = """
<tool id="1">   <test><![CDATA[This is a test.]]></test> </tool>
"""

EXPECTED_FORMATTED_DOCUMENT = """<tool id="1">
    <test><![CDATA[This is a test.]]></test>
</tool>
"""


class TestGalaxyToolFormatServiceClass:
    def test_format_should_return_whole_file_text_edit(self):
        service = GalaxyToolFormatService()
        params = DocumentFormattingParams(
            TextDocumentIdentifier("test"), options=FormattingOptions(tab_size=4, insert_spaces=True),
        )

        actual = service.format(FAKE_UNFORMATTED_DOCUMENT, params)

        assert len(actual) == 1
        assert actual[0].range.start.line == 0
        assert actual[0].range.start.character == 0
        assert actual[0].range.end.line == 3
        assert actual[0].range.end.character == 0

    def test_format_document_returns_expected_format(self):
        service = GalaxyToolFormatService()

        actual = service._format_document(FAKE_UNFORMATTED_DOCUMENT, tabSize=4)

        assert actual == EXPECTED_FORMATTED_DOCUMENT

    def test_format_document_maintains_cdata(self):
        service = GalaxyToolFormatService()

        actual = service._format_document(FAKE_UNFORMATTED_DOCUMENT, tabSize=4)

        assert "<![CDATA[This is a test.]]>" in actual

    def test_format_document_do_not_change_invalid_xml(self):
        service = GalaxyToolFormatService()

        actual = service._format_document(FAKE_INVALID_DOCUMENT, tabSize=4)

        assert actual == FAKE_INVALID_DOCUMENT



"""Galaxy Tools Language Server implementation
"""

from .services.language import GalaxyToolLanguageService
from .features import AUTO_CLOSE_TAGS
from .types import AutoCloseTagResult
from typing import Optional, List
from pygls.server import LanguageServer
from pygls.features import (
    COMPLETION,
    FORMATTING,
    HOVER,
    TEXT_DOCUMENT_DID_CLOSE,
    TEXT_DOCUMENT_DID_OPEN,
    TEXT_DOCUMENT_DID_SAVE,
)
from pygls.types import (
    CompletionList,
    CompletionParams,
    DidCloseTextDocumentParams,
    DidOpenTextDocumentParams,
    DidSaveTextDocumentParams,
    DocumentFormattingParams,
    Hover,
    TextDocumentPositionParams,
    TextEdit,
)

SERVER_NAME = "Galaxy Tools LS"


class GalaxyToolsLanguageServer(LanguageServer):
    """Galaxy Tools Language Server."""

    def __init__(self):
        super().__init__()
        self.service = GalaxyToolLanguageService(SERVER_NAME)


language_server = GalaxyToolsLanguageServer()


@language_server.feature(COMPLETION, trigger_characters=["<", " "])
def completions(server: GalaxyToolsLanguageServer, params: CompletionParams) -> CompletionList:
    """Returns completion items depending on the current document context."""
    document = server.workspace.get_document(params.textDocument.uri)
    return server.service.get_completion(document, params)


@language_server.feature(AUTO_CLOSE_TAGS)
def auto_close_tag(
    server: GalaxyToolsLanguageServer, params: TextDocumentPositionParams
) -> AutoCloseTagResult:
    """Responds to a close tag request to close the currently opened node."""
    document = server.workspace.get_document(params.textDocument.uri)
    return server.service.get_auto_close_tag(document, params)


@language_server.feature(HOVER)
def hover(
    server: GalaxyToolsLanguageServer, params: TextDocumentPositionParams
) -> Optional[Hover]:
    """Displays Markdown documentation for the element under the cursor."""
    document = server.workspace.get_document(params.textDocument.uri)
    return server.service.get_documentation(document, params.position)


@language_server.feature(FORMATTING)
def formatting(
    server: GalaxyToolsLanguageServer, params: DocumentFormattingParams
) -> List[TextEdit]:
    """Formats the whole document using the provided parameters"""
    document = server.workspace.get_document(params.textDocument.uri)
    content = document.source
    return server.service.format_document(content, params)


@language_server.feature(TEXT_DOCUMENT_DID_OPEN)
async def did_open(server: GalaxyToolsLanguageServer, params: DidOpenTextDocumentParams) -> None:
    """Occurs when a new xml document is open."""
    _validate(server, params)


@language_server.feature(TEXT_DOCUMENT_DID_SAVE)
def did_save(server: GalaxyToolsLanguageServer, params: DidSaveTextDocumentParams) -> None:
    """Occurs when the xml document is saved to disk."""
    _validate(server, params)


@language_server.feature(TEXT_DOCUMENT_DID_CLOSE)
def did_close(server: GalaxyToolsLanguageServer, params: DidCloseTextDocumentParams) -> None:
    """Occurs when the xml document is closed."""
    # server.show_message("Xml Document Closed")


def _validate(server: GalaxyToolsLanguageServer, params) -> None:
    """Validates the Galaxy tool and reports any problem found."""
    document = server.workspace.get_document(params.textDocument.uri)
    diagnostics = server.service.get_diagnostics(document)
    server.publish_diagnostics(document.uri, diagnostics)



"""Service in charge of formating XML files following
best practices.
"""

from lxml import etree
from typing import List

from pygls.types import DocumentFormattingParams, Position, Range, TextEdit


class GalaxyToolFormatService:
    """Galaxy tool format service.

    This service manages XML file formatting
    following best practices for Galaxy tools.
    """

    def format(self, content: str, params: DocumentFormattingParams) -> List[TextEdit]:
        """Given the document contents returns the list of TextEdits
        needed to properly layout the document.
        """
        formatted_result = self._format_document(content, params.options.tabSize)

        lines = content.count("\n")
        start = Position(0, 0)
        end = Position(lines + 1, 0)
        return [TextEdit(Range(start, end), formatted_result)]

    def _format_document(self, content: str, tabSize: int) -> str:
        """Formats the whole XML document."""
        try:
            parser = etree.XMLParser(strip_cdata=False)
            xml = etree.fromstring(content, parser=parser)
            spaces = " " * tabSize
            etree.indent(xml, space=spaces)
            result = etree.tostring(xml, pretty_print=True, encoding=str)
            return result
        except etree.XMLSyntaxError:
            return content  # Do not auto-format if there are syntax errors



import pytest

from lxml import etree
from ...services.xsd.constants import TOOL_XSD_FILE
from ...services.xsd.validation import GalaxyToolValidationService
from .sample_data import (
    TEST_TOOL_01_DOCUMENT,
    TEST_MACRO_01_DOCUMENT,
    TEST_INVALID_TOOL_01_DOCUMENT,
    TEST_SYNTAX_ERROR_TOOL_01_DOCUMENT,
    TEST_SYNTAX_ERROR_MACRO_01_DOCUMENT,
)


TEST_SERVER_NAME = "Test Server"


@pytest.fixture(scope="module")
def xsd_schema() -> etree.XMLSchema:
    root = etree.parse(str(TOOL_XSD_FILE))
    schema = etree.XMLSchema(root)
    return schema


class TestGalaxyToolValidationServiceClass:
    def test_validate_document_returns_empty_diagnostics_when_valid(self, xsd_schema: etree.XMLSchema) -> None:
        service = GalaxyToolValidationService(TEST_SERVER_NAME, xsd_schema)

        actual = service.validate_document(TEST_TOOL_01_DOCUMENT)

        assert actual == []

    def test_validate_macro_file_returns_empty_diagnostics_when_valid(self, xsd_schema: etree.XMLSchema) -> None:
        service = GalaxyToolValidationService(TEST_SERVER_NAME, xsd_schema)

        actual = service.validate_document(TEST_MACRO_01_DOCUMENT)

        assert actual == []

    def test_validate_document_returns_diagnostics_when_invalid(self, xsd_schema: etree.XMLSchema) -> None:
        service = GalaxyToolValidationService(TEST_SERVER_NAME, xsd_schema)

        actual = service.validate_document(TEST_INVALID_TOOL_01_DOCUMENT)

        assert len(actual) > 0

    def test_validate_document_returns_diagnostics_when_syntax_error(self, xsd_schema: etree.XMLSchema) -> None:
        service = GalaxyToolValidationService(TEST_SERVER_NAME, xsd_schema)

        actual = service.validate_document(TEST_SYNTAX_ERROR_TOOL_01_DOCUMENT)

        assert len(actual) == 1

    def test_validate_macro_file_returns_diagnostics_when_syntax_error(self, xsd_schema: etree.XMLSchema) -> None:
        service = GalaxyToolValidationService(TEST_SERVER_NAME, xsd_schema)

        actual = service.validate_document(TEST_SYNTAX_ERROR_MACRO_01_DOCUMENT)

        assert len(actual) == 1



"""This module provides a service to determine position context inside an XML document."""

from typing import Optional, List
from io import BytesIO
from pygls.workspace import Document, Position
from pygls.types import Range
from enum import Enum, unique
import xml.sax
import re
from .xsd.types import XsdTree, XsdNode


CDATA_BLOCK_START_SIZE = len("<![CDATA[")
CDATA_BLOCK_END_SIZE = len("]]>")
START_TAG_REGEX = r"<([a-z_]+)[ \n\W]?"
END_TAG_REGEX = r"<\/[\s]*([\w]*)[^>]*>?"
SELF_CLOSED_TAG_REGEX = r"<[\s]*([\w]*)[^>/]*(/[^><]*>?)"
ATTR_KEY_VALUE_REGEX = r" ([a-z_]*)=\"([\w. ]*)[\"]?"
TAG_GROUP = 1
ATTR_KEY_GROUP = 1
ATTR_VALUE_GROUP = 2
END_SELF_TAG_GROUP = 2


@unique
class ContextTokenType(Enum):
    """Supported types of tokens that can be found in a document."""

    UNKNOWN = 1
    TAG = 2
    ATTRIBUTE_KEY = 3
    ATTRIBUTE_VALUE = 4


class XmlContext:
    """Represents the context at a given XML document position.

    It provides information about the token under the cursor and
    the XSD node definition associated.
    """

    def __init__(
        self,
        document_line: str = "",
        position: Position = None,
        is_empty: bool = False,
        token_name: str = None,
        token_type: ContextTokenType = ContextTokenType.UNKNOWN,
    ):
        self.document_line = document_line
        self.target_position = position
        self.is_empty = is_empty
        self.token_name = token_name
        self.token_type = token_type
        self.token_range: Range = None
        self.attr_name = None
        self.attr_list: List[str] = []
        self.node: Optional[XsdNode] = None
        self.is_node_content: bool = False
        self.is_closing_tag: bool = False
        self.node_stack: List[str] = []

    def is_tag(self) -> bool:
        """Indicates if the token in context is a tag"""
        return self.token_type == ContextTokenType.TAG

    def is_attribute_key(self) -> bool:
        """Indicates if the token in context is an attribute key"""
        return self.token_type == ContextTokenType.ATTRIBUTE_KEY

    def is_attribute_value(self) -> bool:
        """Indicates if the token in context is an attribute value"""
        return self.token_type == ContextTokenType.ATTRIBUTE_VALUE

    def update_node_definition(self, xsd_tree: XsdTree) -> None:
        """Updates the associated node in the context using the given XSD tree definition.

        Args:
            xsd_tree (XsdTree): The XSD tree definition.
        """
        self.node = xsd_tree.find_node_by_stack(self.node_stack)
        if not self.node:
            self.node = xsd_tree.root


class XmlContextService:
    """This service provides information about the XML context at
    a specific position of the document.
    """

    def __init__(self, xsd_tree: XsdTree):
        self.xsd_tree = xsd_tree

    def get_xml_context(self, document: Document, position: Position) -> XmlContext:
        """Gets the XML context at a given offset inside the document.

        Args:
            document (Document): The current document.
            position (Position): The position inside de document.

        Returns:
            XmlContext: The resulting context with the current node
            definition and other information. If the context can not be
            determined, the default context with no information is returned.
        """
        parser = XmlContextParser()
        context = parser.parse(document, position)
        context.update_node_definition(self.xsd_tree)
        return context


class XmlContextParser:
    """This class parses an XML document until it finds the token at a given
    position (line and character) and returns information about it.

    The parser will try it's best to be tolerant to imcomplete XML documents
    in order to be able to extract the token at position.
    """

    def __init__(self):
        self._document: Document = None

    def parse(self, document: Document, position: Position) -> XmlContext:
        """Parses the given document until it finds the token at the given position and
        returns context information about this token.

        Args:
            document (Document): The XML document to parse.
            position (Position): The position inside the document to get the context.

        Returns:
            XmlContext: The context information at the given position.
        """
        self._document = document

        if self.is_empty_document(document):
            return XmlContext(is_empty=True)

        context_line = document.lines[position.line]
        context = XmlContext(document_line=context_line, position=position)

        source = BytesIO(document.source.encode())
        reader = self._build_xml_reader(context)

        try:
            reader.parse(source)
        except ContextFoundException:
            pass  # All is good, we got what we needed

        return context

    def _build_xml_reader(self, context: XmlContext) -> xml.sax.xmlreader.XMLReader:
        reader = xml.sax.make_parser()
        reader.setFeature(xml.sax.handler.feature_namespaces, False)
        reader.setFeature(xml.sax.handler.feature_validation, False)
        locator = xml.sax.expatreader.ExpatLocator(reader)
        saxhandler = ContextBuilderHandler(locator, context)
        reader.setProperty(xml.sax.handler.property_lexical_handler, saxhandler)
        reader.setContentHandler(saxhandler)
        reader.setErrorHandler(ContextParseErrorHandler(context))
        return reader

    @staticmethod
    def is_empty_document(document: Document) -> bool:
        """Determines if the given XML document is an empty document.

        Args:
            document (Document): The document to check.

        Returns:
            bool: True if the document is empty and False otherwise.
        """
        xml_content = document.source.strip()
        return not xml_content or len(xml_content) < 2


class ContextBuilderHandler(xml.sax.ContentHandler):
    """SAX ContentHandler that tries to compose the context information using the
    different SAX events.
    """

    def __init__(self, locator, context: XmlContext):
        xml.sax.ContentHandler.__init__(self)
        self._loc = locator
        self._context: XmlContext = context
        self.setDocumentLocator(self._loc)

    def startDocument(self):
        pass

    def endDocument(self):
        pass

    def startElement(self, tag, attributes):
        self._context.node_stack.append(tag)
        current_position = self.get_current_position()
        if current_position.line == self._context.target_position.line:
            self._build_context_from_element_line(current_position.character, tag, attributes)

    def endElement(self, tag):
        current_position = self.get_current_position()
        if current_position.line < self._context.target_position.line:
            self._context.node_stack.pop(-1)
        elif current_position.line == self._context.target_position.line:
            match_close = re.search(f"<[\\s]*/[\\s]*{tag}[\\s]*>", self._context.document_line)
            tag_start = current_position.character
            if match_close and tag_start <= self._context.target_position.character <= match_close.end(0):
                self._context.is_closing_tag = True
                self._context.token_type = ContextTokenType.TAG
                self._context.token_name = tag
                self._context.token_range = Range(
                    Position(line=current_position.line, character=tag_start),
                    Position(line=current_position.line, character=match_close.end(0)),
                )
                raise ContextFoundException()
            if (
                match_close
                and self._context.target_position.character > match_close.end(0)
                or not match_close
                and self._context.target_position.character >= current_position.character
            ):
                self._context.node_stack.pop(-1)

            match_close = re.search(SELF_CLOSED_TAG_REGEX, self._context.document_line)
            if match_close:
                tag = match_close.group(TAG_GROUP)
                tag_start = match_close.start(END_SELF_TAG_GROUP)
                tag_end = match_close.end(END_SELF_TAG_GROUP)
                if tag_start < self._context.target_position.character < tag_end:
                    self._context.is_closing_tag = True
                    raise ContextFoundException()

            if current_position.character > self._context.target_position.character:
                raise ContextFoundException()

    def characters(self, content):
        if self._is_target_at_current_element(len(content)):
            self._context.is_node_content = True
            raise ContextFoundException()

    def comment(self, content):
        pass

    def startDTD(self, name, public_id, system_id):
        pass

    def endDTD(self):
        pass

    def startEntity(self, name):
        pass

    def endEntity(self, name):
        pass

    def startCDATA(self):
        if self._is_target_at_current_element(CDATA_BLOCK_START_SIZE):
            self._context.is_node_content = True
            raise ContextFoundException()

    def endCDATA(self):
        if self._is_target_at_current_element(CDATA_BLOCK_END_SIZE):
            self._context.is_node_content = True
            raise ContextFoundException()

    def get_current_position(self) -> Position:
        """Gets the current position inside the document where the parser is.

        Returns:
            Position: The current position inside the document.
        """
        return Position(line=self._loc.getLineNumber() - 1, character=self._loc.getColumnNumber())

    def _is_target_at_current_element(self, element_size: int) -> bool:
        current_position = self.get_current_position()
        if current_position.line == self._context.target_position.line:
            element_start = current_position.character
            content_end = current_position.character + element_size
            return element_start <= self._context.target_position.character <= content_end
        elif current_position.line > self._context.target_position.line:
            raise ContextFoundException()
        return False

    def _build_context_from_element_line(self, start_position: int, tag: str, attributes):
        self._context.attr_list = list(attributes.keys())
        target_offset = self._context.target_position.character
        tag_offset = start_position + len(tag)
        is_on_tag = start_position <= target_offset <= tag_offset
        if is_on_tag:
            self._context.token_type = ContextTokenType.TAG
            self._context.token_name = tag
            self._context.token_range = Range(
                Position(line=self._context.target_position.line, character=start_position + 1),
                Position(line=self._context.target_position.line, character=tag_offset + 1),
            )
            raise ContextFoundException()

        accum = tag_offset + 1  # +1 for '<'
        for attr_name, attr_value in attributes.items():
            attr_start = accum + 1  # +1 for ' '
            attr_end = attr_start + len(attr_name)
            is_on_attr = attr_start <= target_offset <= attr_end
            if is_on_attr:
                self._context.token_type = ContextTokenType.ATTRIBUTE_KEY
                self._context.token_name = attr_name
                self._context.token_range = Range(
                    Position(line=self._context.target_position.line, character=attr_start),
                    Position(line=self._context.target_position.line, character=attr_end),
                )
                raise ContextFoundException()

            attr_value_start = attr_end + 2  # +2 for '=' and '\"'
            attr_value_end = attr_value_start + len(attr_value)
            is_on_attr_value = attr_value_start <= target_offset <= attr_value_end
            if is_on_attr_value:
                self._context.token_type = ContextTokenType.ATTRIBUTE_VALUE
                self._context.token_name = attr_value
                self._context.attr_name = attr_name
                self._context.token_range = Range(
                    Position(line=self._context.target_position.line, character=attr_value_start),
                    Position(line=self._context.target_position.line, character=attr_value_end),
                )
                raise ContextFoundException()
            accum = attr_value_end + 1  # closing \"


class ContextParseErrorHandler(xml.sax.ErrorHandler):
    """SAX ErrorHandler that tries it's best to recover context information when a fatal error
    ocurred during parsing.
    """

    def __init__(self, context: XmlContext):
        xml.sax.ErrorHandler.__init__(self)
        self._context = context

    def fatalError(self, exception: xml.sax.SAXParseException):
        position = self.get_position(exception)
        self._try_recover_context_from_exception(position.character)

    def get_position(self, exception: xml.sax.SAXParseException) -> Position:
        return Position(line=exception.getLineNumber() - 1, character=exception.getColumnNumber())

    def _try_recover_context_from_exception(self, start_offset: int) -> None:
        target_offset = self._context.target_position.character
        self._try_get_attribute_context_at_line_position(target_offset)
        self._try_get_tag_context_at_line_position(target_offset)

    def _try_get_tag_context_at_line_position(self, target_offset: int) -> None:
        tag_matches = re.finditer(START_TAG_REGEX, self._context.document_line, re.DOTALL)
        if tag_matches:
            for match in tag_matches:
                tag = match.group(TAG_GROUP)
                tag_start = match.start(TAG_GROUP)
                tag_end = match.end(TAG_GROUP)
                if tag_start <= target_offset <= tag_end:
                    self._context.token_type = ContextTokenType.TAG
                    self._context.token_name = tag
                    self._context.token_range = Range(
                        Position(line=self._context.target_position.line, character=tag_start),
                        Position(line=self._context.target_position.line, character=tag_end),
                    )
                    if tag not in self._context.node_stack:
                        self._context.node_stack.append(tag)
                    raise ContextFoundException()

                if (
                    target_offset > tag_end
                    and tag not in self._context.node_stack
                    and not self._is_tag_closed_before_offset(tag, target_offset)
                ):
                    self._context.node_stack.append(tag)

        match_close = re.search(END_TAG_REGEX, self._context.document_line)
        if match_close:
            tag = match_close.group(TAG_GROUP)
            tag_start = match_close.start(TAG_GROUP)
            tag_end = match_close.end(TAG_GROUP)
            if tag_start <= self._context.target_position.character <= tag_end:
                self._raise_close_context_found(tag, tag_start, tag_end)
        match_close = re.search(SELF_CLOSED_TAG_REGEX, self._context.document_line)
        if match_close:
            tag = match_close.group(TAG_GROUP)
            tag_start = match_close.start(END_SELF_TAG_GROUP)
            tag_end = match_close.end(END_SELF_TAG_GROUP)
            if tag_start < self._context.target_position.character < tag_end:
                self._raise_close_context_found(tag, tag_start, tag_end)

    def _raise_close_context_found(self, tag, tag_start, tag_end):
        self._context.is_closing_tag = True
        self._context.token_type = ContextTokenType.TAG
        self._context.token_name = tag
        self._context.token_range = Range(
            Position(line=self._context.target_position.line, character=tag_start),
            Position(line=self._context.target_position.line, character=tag_end),
        )
        if tag not in self._context.node_stack:
            self._context.node_stack.append(tag)
        raise ContextFoundException()

    def _is_tag_closed_before_offset(self, tag: str, offset: int) -> bool:
        match_close = re.search(f"<[\\s]*/[\\s]*{tag}[\\s]*>", self._context.document_line)
        if match_close and offset >= match_close.end(0):
            return True
        match_self_close = re.search(f"<{tag}[^>]*/>", self._context.document_line)
        return match_self_close and offset >= match_self_close.end(0)

    def _try_get_attribute_context_at_line_position(self, target_offset):
        attribute_matches = re.finditer(ATTR_KEY_VALUE_REGEX, self._context.document_line, re.DOTALL)
        self._context.attr_list = []
        for match in attribute_matches:
            self._context.attr_list.append(match.group(ATTR_KEY_GROUP))
            if match.start(ATTR_KEY_GROUP) <= target_offset <= match.end(ATTR_KEY_GROUP):
                self._context.token_type = ContextTokenType.ATTRIBUTE_KEY
                self._context.token_name = match.group(ATTR_KEY_GROUP)
                self._context.token_range = Range(
                    Position(line=self._context.target_position.line, character=match.start(ATTR_KEY_GROUP),),
                    Position(line=self._context.target_position.line, character=match.end(ATTR_KEY_GROUP),),
                )
                raise ContextFoundException()

            if match.start(ATTR_VALUE_GROUP) <= target_offset <= match.end(ATTR_VALUE_GROUP):
                self._context.token_type = ContextTokenType.ATTRIBUTE_VALUE
                self._context.token_name = match.group(ATTR_VALUE_GROUP)
                self._context.attr_name = match.group(ATTR_KEY_GROUP)
                self._context.token_range = Range(
                    Position(line=self._context.target_position.line, character=match.start(ATTR_VALUE_GROUP),),
                    Position(line=self._context.target_position.line, character=match.end(ATTR_VALUE_GROUP),),
                )
                raise ContextFoundException()


class ContextFoundException(Exception):
    """When this exception is raised, it indicates that the necessary
    information for the context is already retrieved, so, additional
    parsing of the file can be avoided.
    """

    pass
